#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/io.h>
#include <sys/kd.h>
#include <linux/fb.h>

int fd;
int root;
struct fb_fix_screeninfo fixinfo;
struct fb_var_screeninfo varinfo;

#define STEP 32

int wait_fb(void) {
	struct fb_vblank blank;

	if (ioctl(fd, FBIOGET_VBLANK, &blank) != 0) {
		return -1;
	}

	if ((blank.flags & FB_VBLANK_HAVE_COUNT) != 0) {
		unsigned start = blank.count;
		while (start == blank.count) {
			if (ioctl(fd, FBIOGET_VBLANK, &blank) != 0) {
				return -1;
			}
		}
	} else if ((blank.flags & FB_VBLANK_HAVE_VSYNC) != 0) {
		while ((blank.flags & FB_VBLANK_VSYNCING) == 0) {
			if (ioctl(fd, FBIOGET_VBLANK, &blank) != 0) {
				return -1;
			}
		}
	} else if ((blank.flags & FB_VBLANK_HAVE_VBLANK) != 0) {
		while ((blank.flags & FB_VBLANK_VBLANKING) == 0) {
			if (ioctl(fd, FBIOGET_VBLANK, &blank) != 0) {
				return -1;
			}
		}
	} else {
		return -1;
	}

	return 0;
}

int wait_vga(void)
{
	const unsigned limit = 100000;
	unsigned counter;

	if (!root)
		return -1;

	counter = 0;

	while ((inb(0x3da) & 0x8) != 0 && counter < limit)
		++counter;

	while ((inb(0x3da) & 0x8) == 0 && counter < limit)
		++counter;

	if (counter > limit) {
		root = 0;
	}

	return 0;
}

void wait_some(void)
{
	struct timespec req;

	req.tv_sec = 0;
	req.tv_nsec = 10 * 1000000;

	nanosleep(&req, 0);
}

void delay(void)
{
	if (wait_fb() == 0) {
		return;
	}

	if (wait_vga() == 0) {
		return;
	}

	wait_some();
}

unsigned red_mask;
unsigned green_mask;
unsigned blue_mask;

unsigned red_index[256];
unsigned green_index[256];
unsigned blue_index[256];

void fadeout_precompute(unsigned char* d, unsigned p)
{
	unsigned r, g, b;

	r = (p & red_mask) >> varinfo.red.offset;
	g = (p & green_mask) >> varinfo.green.offset;
	b = (p & blue_mask) >> varinfo.blue.offset;

	d[0] = r;
	d[1] = g;
	d[2] = b;
}

void fadeout_setup(unsigned val, unsigned base)
{
	unsigned i;

	for(i=0;i<256;++i) {
		red_index[i] = (((i * val) / base) << varinfo.red.offset) & red_mask;
		green_index[i] = (((i * val) / base) << varinfo.green.offset) & green_mask;
		blue_index[i] = (((i * val) / base) << varinfo.blue.offset) & blue_mask;
	}
}

int fadeout_truecolor(void)
{
	unsigned char* ptr;
	unsigned char* src;
	unsigned char* buf;
	unsigned pixel;
	unsigned scanline;
	int i;
	unsigned x,y;

	red_mask = ((1 << varinfo.red.length) - 1) << varinfo.red.offset;
	green_mask = ((1 << varinfo.green.length) - 1) << varinfo.green.offset;
	blue_mask = ((1 << varinfo.blue.length) - 1) << varinfo.blue.offset;

	pixel = (varinfo.bits_per_pixel + 7) / 8;

	if (pixel != 2 && pixel != 4)
		return -1;

	scanline = varinfo.xres * pixel;

	src = malloc(varinfo.yres * varinfo.xres * 3);
	if (!src) {
		return -1;
	}
	buf = malloc(varinfo.yres * scanline);
	if (!buf) {
		return -1;
	}

	ptr = mmap(0,
		fixinfo.smem_len,
		PROT_READ | PROT_WRITE,
		MAP_SHARED | MAP_FIXED,
		fd,
		0
	);
	if (ptr == MAP_FAILED) {
		return -1;
	}

	for(y=0;y<varinfo.yres;++y) {
		unsigned char* rptr = ptr + y * fixinfo.line_length;
		unsigned char* rbuf = buf + y * scanline;
		memcpy(rbuf, rptr, scanline);
	}

	if (pixel == 2) {
		for(y=0;y<varinfo.yres;++y) {
			unsigned char* rsrc = src + y * varinfo.xres * 3;
			unsigned short* rbuf = (unsigned short*)(buf + y * scanline);
			for(x=0;x<varinfo.xres;++x) {
				fadeout_precompute(rsrc + x*3, rbuf[x]);
			}
		}
	} else {
		for(y=0;y<varinfo.yres;++y) {
			unsigned char* rsrc = src + y * varinfo.xres * 3;
			unsigned int* rbuf = (unsigned int*)(buf + y * scanline);
			for(x=0;x<varinfo.xres;++x) {
				fadeout_precompute(rsrc + x*3, rbuf[x]);
			}
		}
	}

	for(i=STEP-1;i>=0;--i) {
		fadeout_setup(i,STEP);

		if (pixel == 2) {
			for(y=0;y<varinfo.yres;++y) {
				unsigned char* rsrc = src + y * varinfo.xres * 3;
				unsigned short* rbuf = (unsigned short*)(buf + y * scanline);
				for(x=0;x<varinfo.xres;++x) {
					*rbuf = red_index[rsrc[0]] | green_index[rsrc[1]] | blue_index[rsrc[2]];
					rsrc += 3;
					rbuf += 1;
				}
			}
		} else {
			for(y=0;y<varinfo.yres;++y) {
				unsigned char* rsrc = src + y * varinfo.xres * 3;
				unsigned int* rbuf = (unsigned int*)(buf + y * scanline);
				for(x=0;x<varinfo.xres;++x) {
					*rbuf = red_index[rsrc[0]] | green_index[rsrc[1]] | blue_index[rsrc[2]];
					rsrc += 3;
					rbuf += 1;
				}
			}
		}

		delay();

		for(y=0;y<varinfo.yres;++y) {
			unsigned char* rptr = ptr + y * fixinfo.line_length;
			unsigned char* rbuf = buf + y * scanline;
			memcpy(rptr, rbuf, scanline);
		}
	}

	munmap(ptr, fixinfo.smem_len);
	return 0;
}

int clear(void)
{
	unsigned char* ptr;

	ptr = mmap(0,
		fixinfo.smem_len,
		PROT_READ | PROT_WRITE,
		MAP_SHARED | MAP_FIXED,
		fd,
		0
	);
	if (ptr == MAP_FAILED) {
		return -1;
	}

	memset(ptr, 0, fixinfo.line_length * varinfo.yres);

	munmap(ptr, fixinfo.smem_len);

	return 0;
}

int main(void) {
	const char* fb;

	if (iopl(3) == 0) {
		root = 1;
	} else {
		root = 0;
	}

	fb = getenv("FRAMEBUFFER");
	if (fb && fb[0]) {
		fd = open(fb, O_RDWR);
	} else {
		fb = "/dev/fb0";
		fd = open(fb, O_RDWR);
		if (fd < 0 && errno == ENOENT) {
			fb = "/dev/fb/0";
			fd = open(fb, O_RDWR);
		}
	}
	if (fd < 0) {
		exit(1);
	}

	memset(&fixinfo, 0, sizeof(fixinfo));

	if (ioctl(fd, FBIOGET_FSCREENINFO, &fixinfo) != 0) {
		exit(1);
	}

	memset(&varinfo, 0, sizeof(varinfo));

	if (ioctl(fd, FBIOGET_VSCREENINFO, &varinfo) != 0) {
		exit(1);
	}

	/* hack for a bug on the kernel Radeon driver (2.4.22) */
	if (varinfo.red.offset == varinfo.blue.offset) {
		varinfo.red.offset = 10;
		varinfo.blue.offset = 0;
	}

	fadeout_truecolor();

	clear();

	close(fd);

	return 0;
}

